/*
 *   How to compile?
 *      gcc -Wall -o fft2 fft2.c -lm -lfftw -lcfitsio
 *
 *   How to execute?
 *      fft2 [input image] [output image]
 *
 */

#include<stdio.h>
#include<math.h>
#include<stdlib.h>
#include<string.h>
#include<fftw.h>
#include<fitsio.h>


/* FFT */
void fft2(fftw_complex *input, fftw_complex *output, long w, long h)
{
  int i, j, hw, hh;
  fftw_complex *pa, *pb, *pc, *pd;
  fftwnd_plan  plan;
  float re, im;
  
  plan = fftw2d_create_plan(w, h, FFTW_FORWARD, FFTW_ESTIMATE);
  fftwnd_one(plan, input, output);

  /* fft shift 
   *
   *    Freq. dom.
   *  +-----+-----+         +-----+-----+
   *  |     |     |         |     |     |
   *  |  A  |  B  |         |  D  |  C  |
   *  |     |     |         |     |     |
   *  +-----+-----+   -->   +-----+-----+
   *  |     |     |         |     |     |
   *  |  C  |  D  |         |  B  |  A  |
   *  |     |     |         |     |     |
   *  +-----+-----+         +-----+-----+
   *
   */
  hw = w / 2;
  hh = h / 2;
  for (j = 0; j < hh; j ++) {
    pa = &output[j*w]; 
    pb = pa+hw;
    pc = pa+w*hh; 
    pd = pc + hw;
    for (i = 0; i < hw; i ++) {
      re = pa->re; im = pa->im;
      pa->re = pd->re; pa->im = pd->im;
      pd->re = re; pd->im = im;
      re = pb->re; im = pb->im;
      pb->re = pc->re; pb->im = pc->im;
      pc->re = re; pc->im = im;
      pa++;
      pb++;
      pc++;
      pd++;
    }
  }
  fftwnd_destroy_plan(plan);
}

/* inverse fft */
void ifft2(fftw_complex *input, fftw_complex *output, long w, long h)
{
  int i, j, hw, hh;
  fftw_complex *pa, *pb, *pc, *pd;
  fftwnd_plan  plan;
  float re, im, size;

  /* fft shift 
   *
   *    Freq. dom.
   *  +-----+-----+         +-----+-----+
   *  |     |     |         |     |     |
   *  |  D  |  C  |         |  A  |  B  |
   *  |     |     |         |     |     |
   *  +-----+-----+   -->   +-----+-----+
   *  |     |     |         |     |     |
   *  |  B  |  A  |         |  C  |  D  |
   *  |     |     |         |     |     |
   *  +-----+-----+         +-----+-----+
   *
   */
  hw = w / 2;
  hh = h / 2;
  size = w*h;
  for (j = 0; j < hh; j ++) {
    pa = &input[j*w]; 
    pb = pa+hw;
    pc = pa+w*hh; 
    pd = pc + hw;
    for (i = 0; i < hw; i ++) {
      re = pa->re; im = pa->im;
      pa->re = pd->re/size; pa->im = pd->im/size;
      pd->re = re/size; pd->im = im/size;
      re = pb->re; im = pb->im;
      pb->re = pc->re/size; pb->im = pc->im/size;
      pc->re = re/size; pc->im = im/size;
      pa++;
      pb++;
      pc++;
      pd++;
    }
  }

  plan = fftw2d_create_plan(w, h, FFTW_BACKWARD, FFTW_ESTIMATE);
  fftwnd_one(plan, input, output);

  fftwnd_destroy_plan(plan);
}

int writeimg (char *fname, double *input, long imwidth, long imheight){
  fitsfile *fptr;
  int status = 0;
  int naxis=2;
  long fpixel[2] = {1, 1};
  long npixels;
  long naxes[2];
  long i,j;
  double *pix;

  naxes[0] = imwidth;
  naxes[1] = imheight;

  npixels = naxes[0] * naxes[1];

  fits_create_file(&fptr, fname, &status);
  if(status)
    return(status);
  fits_create_img(fptr, DOUBLE_IMG, naxis, naxes, &status);
  if(status)
    return(status);

  pix = (double *) malloc(naxes[0] * sizeof(double));

  fpixel[0] = 1;
  
  for(j=1; j<=naxes[1]; j++){
    fpixel[1] = j;
    for(i=0; i<naxes[0]; i++){
      pix[i] = input[(j-1)*naxes[0] + i];
    }
    fits_write_pix(fptr, TDOUBLE, fpixel, naxes[0], pix, &status);
    if(status)
      return(status);
  }
  
  fits_close_file(fptr, &status);
  free(pix);
  
  return(status);
}

int main(int argc, char *argv[]){
  fitsfile *fptr1,*fptr2;
  int status = 0;
  int dim;
  float mag_fac;
  int binpix;
  long i,j,l,m,imwidth, imheight, imwidth2, imheight2;
  long naxes[2] = {1, 1}, fpixel[2] = {1, 1};
  long npixels;
  double *pix, *inpix, *phasepix, *outpix, *outpix2;
  double flux = 0;
  
  fftw_complex *infft, *outfft;

  if(argc != 5){
    printf("Usage: fft2 [input image] [phase image] [magnification factor] [output image]\n");
    return(1);
  }

  fits_open_file(&fptr1, argv[1], READONLY, &status);
  fits_get_img_dim(fptr1, &dim, &status);
  fits_get_img_size(fptr1, 2, naxes, &status);
  
  fits_open_file(&fptr2, argv[2], READONLY, &status);
  fits_get_img_dim(fptr2, &dim, &status);
  fits_get_img_size(fptr2, 2, naxes, &status);

  if(status){
    fits_report_error(stderr, status);
    return (status);
  }
  
  imwidth = naxes[0];
  imheight = naxes[1];
  npixels = imwidth * imheight;
  
  pix = (double *) malloc(imwidth * sizeof(double));
  inpix = (double *) malloc(npixels * sizeof(double));
  phasepix = (double *) malloc(npixels * sizeof(double));

  if(pix == NULL || inpix == NULL || phasepix == NULL){
    printf("Memory allocation error\n");
    return (1);
  }

  fpixel[0] = 1;

  for(j=1; j<=imheight; j++){
    fpixel[1] = j;
    if(fits_read_pix(fptr1, TDOUBLE, fpixel, imwidth, 0, pix, 0, &status))
      break;

    for(i=0; i < imwidth; i++){
      inpix[(j-1)*imwidth + i] = pix[i];
    }
  }

  if(status){    
    fits_report_error(stderr, status);
    return (status);
  }

  
  fpixel[0] = 1;

  for(j=1; j<=imheight; j++){
    fpixel[1] = j;
    if(fits_read_pix(fptr2, TDOUBLE, fpixel, imwidth, 0, pix, 0, &status))
      break;

    for(i=0; i < imwidth; i++){
      phasepix[(j-1)*imwidth + i] = pix[i];
    }
  }

  if(status){    
    fits_report_error(stderr, status);
    return (status);
  }  

  infft = fftw_malloc(sizeof(fftw_complex)*npixels);
  outfft = fftw_malloc(sizeof(fftw_complex)*npixels);

  for(j=0;j<imheight;j++){
    for(i=0;i<imwidth;i++){
      infft[j*imwidth+i].re = inpix[j*imwidth+i] * cos(phasepix[j*imwidth+i]);
      infft[j*imwidth+i].im = inpix[j*imwidth+i] * sin(phasepix[j*imwidth+i]);
    }
  }

  fft2(infft, outfft, imwidth, imheight);
  outpix = (double *) malloc(npixels * sizeof(double));
  
  for(j=0;j<imheight;j++){
    for(i=0;i<imwidth;i++){
      flux += outfft[j*imwidth + i].re * outfft[j*imwidth + i].re + outfft[j*imwidth + i].im * outfft[j*imwidth + i].im;
    }
  }

  for(j=0;j<imheight;j++){
    for(i=0;i<imwidth;i++){
      outpix[j*imwidth + i] = (outfft[j*imwidth + i].re * outfft[j*imwidth + i].re + outfft[j*imwidth + i].im * outfft[j*imwidth + i].im)/flux;
    }
  }

  /* magnification factor */
  mag_fac = atof(argv[3]);

  imwidth2 = imwidth * mag_fac;
  imheight2 = imheight * mag_fac;
  
  outpix2 = (double *) malloc(imwidth2*imheight2*sizeof(double));
  
  for(j=0;j<imheight2;j++){
    for(i=0;i<imwidth2;i++){
      outpix2[j*imwidth2+i] = 0;
    }
  }

  binpix = (int) (1.0 / mag_fac + 0.5);

  for(j=0;j<imheight2;j++){
    for(i=0;i<imwidth2;i++){
      for(m=0;m<binpix;m++){
	for(l=0;l<binpix;l++){
	  outpix2[j*imwidth2+i] += outpix[(j*binpix+m)*imwidth+(i*binpix+l)];
	}
      }
    }
  }

  status = writeimg(argv[4], outpix2, imwidth2, imheight2);
  fits_report_error(stderr, status);
  
  free(infft);
  free(outfft);
  free(pix);
  free(inpix);
  free(outpix);
  free(outpix2);
  free(phasepix);

  return (0);
}
